#version 120
#extension GL_EXT_texture_array : enable

// Varying.
varying vec4 vPos;
varying vec3 worldNormal;
varying vec3 worldPos;

// Global directional light uniforms.
uniform vec4 dirLightDir;
uniform vec4 dirLightColor;
uniform vec4 dirLightAmbient;
uniform vec4 dirShadowColor;

// Misc uniforms.
uniform vec3 camPos;
uniform mat4 obj2World;
uniform mat4 world2Cam;

uniform int isParticle;
uniform int particleGlow;
uniform int doColorMultiply;

// Surface calculations, including specular power.
varying vec2 texCoord;
vec4 viewDelta;
float specular;
float NdotL;
vec3 reflectVec;

uniform sampler2D tex;
void calculateSurface(vec4 color, inout vec4 albedo)
{
   viewDelta.xyz = worldPos - camPos;
   viewDelta.w   = length(viewDelta.xyz);
   viewDelta.xyz = -normalize(viewDelta.xyz);

   vec4 texAlbedo = texture2D(tex, texCoord);
   albedo.rgb = mix(color.rgb, texAlbedo.rgb, texAlbedo.a);

   if(doColorMultiply == 1)
      albedo *= gl_Color;

   albedo.a = color.a;

   NdotL = max(dot(worldNormal, dirLightDir.xyz), 0.0f);
   reflectVec = normalize(reflect(-dirLightDir.xyz, worldNormal));
   specular = pow(max(dot(reflectVec, viewDelta.xyz), 0.0f), 12.0f) * length(texAlbedo.rgb);
   
   //albedo.rgb = normalize(viewDelta.xyz);
}

// Fogging.
uniform vec4 fogBaseColor;
uniform vec4 fogConsts;
uniform sampler2D fogTex;
varying vec2 fogCoords;
void applyFog(inout vec4 albedo)
{
   // Calculate fog.
   vec4 fogColor = texture2D(fogTex, fogCoords) * fogBaseColor;
   
   // Blend it.
   albedo = mix(albedo, fogColor, fogColor.a);
}

// Shadowing
uniform vec4 far_d;
uniform vec2 texSize; // x - size, y - 1/size
uniform vec4 zScale;
uniform int shadowSplitCount;
void calculateShadowCoords(inout vec4 shadow_coordA, inout vec4 shadow_coordB, out float blend)
{
   int index = 3;
   float fudgeKey = 0.1f;
   float fudgeFactorA = 0.0f;
   float fudgeFactorB = 0.0f;
   fudgeFactorA = 4.0f * fudgeKey / zScale.w;
   fudgeFactorB = 4.0f * fudgeKey / zScale.w;
   blend = 0.0f;
   
   // find the appropriate depth map to look up in based on the depth of this fragment
   if(vPos.y < far_d.x)
   {
      index = 0;
      blend = clamp( (vPos.y - (far_d.x * 0.9f)) * 0.25f , 0.0f, 1.0f); 
      fudgeFactorA =        fudgeKey / zScale.x;
      fudgeFactorB = 2.0f * fudgeKey / zScale.y;
   }
   else if(vPos.y < far_d.y)
   {
      index = 1;
      blend = clamp( (vPos.y - (far_d.y * 0.9f)) * 0.25f , 0.0f, 1.0f); 
      fudgeFactorA = 2.0f * fudgeKey / zScale.y;
      fudgeFactorB = 3.0f * fudgeKey / zScale.z;
   }
   else if(vPos.y < far_d.z)
   {
      index = 2;
      blend = clamp( (vPos.y - (far_d.z * 0.9f)) * 0.25f , 0.0f, 1.0f); 
      fudgeFactorA = 3.0f * fudgeKey / zScale.z;
      fudgeFactorB = 4.0f * fudgeKey / zScale.w;
   }
   
   // transform this fragment's position from view space to scaled light clip space
   // such that the xy coordinates are in [0;1]
   // note there is no need to divide by w for orthogonal light sources
   shadow_coordA   = gl_TextureMatrix[index]*vPos;
   shadow_coordA.w = shadow_coordA.z - fudgeFactorA; // Figure the input coordinate for PCF sampling if appropriate.      
   shadow_coordA.z = float(index);                   // Encode the layer to sample.

   //don't have to set second shadow coord if we're not blending
   if(blend > 0.0f)
   {
      shadow_coordB   = gl_TextureMatrix[index + 1]*vPos;
      shadow_coordB.w = shadow_coordB.z - fudgeFactorB;
      shadow_coordB.z = float(index + 1);      
   }
}

vec4 colorRGBtoYCbCr(vec4 colorRGB)
{

   vec4 colorYCbCr = vec4(0.0f, 0.0f, 0.0f, 0.0f);	
	
   float wRed = 0.299f;
   float wBlue = 0.114f;
   float wGreen = 1.0f - wRed - wBlue;
   float uMax = 0.436f;
   float vMax = 0.615f;

   colorYCbCr.x = 0.299f * colorRGB.r + 0.557f * colorRGB.g + 0.114f * colorRGB.b;
   colorYCbCr.y = -0.168736f * colorRGB.r + (-0.331264f * colorRGB.g) + (0.5f * colorRGB.b);
   colorYCbCr.z = 0.5f * colorRGB.r + (-0.418688f * colorRGB.g) + (-0.081312f * colorRGB.b);
   colorYCbCr.w = colorRGB.a;
	
   return colorYCbCr;
}

vec4 colorYCbCrtoRGB(vec4 colorYCbCr)
{

   vec4 colorRGB = vec4(0.0f, 0.0f, 0.0f, 0.0f);	

   colorRGB.r = colorYCbCr.x + (colorYCbCr.z * 1.402f);
   colorRGB.g = colorYCbCr.x + (-0.344136f * colorYCbCr.y) + (-0.714136 * colorYCbCr.z);
   colorRGB.b = colorYCbCr.x + (colorYCbCr.y * 1.772f);
   colorRGB.a = colorYCbCr.w;

   return colorRGB;
}

// Point lighting
const int numlights = 8;
uniform int pointLightCount;
uniform vec4 pointLightPos[numlights];
uniform vec4 pointLightColor[numlights];
uniform float pointLightRadius[numlights];
vec4 accumulatePointLights(vec4 baseLightColor)
{
   vec4 pointLightTotal = vec4(0.0f, 0.0f, 0.0f, 0.0f);
   baseLightColor = colorRGBtoYCbCr(baseLightColor);

   // Calculate effects of the up-to-8-point-lights.
   for(int i = 0; i < pointLightCount; i++)
   {
      vec3 lightDelta = worldPos.xyz - pointLightPos[i].xyz;
      float lightDist = length(lightDelta);
      float lightDot = max(dot(-normalize(lightDelta), worldNormal), 0.0f);
      if(isParticle == 1)
         lightDot = 1.0f;
      if(lightDist < pointLightRadius[i] * 1.5f)
      {
         vec4 pointLightColorYCbCr = colorRGBtoYCbCr(pointLightColor[i]);

         float ratio = 1.0f - (lightDist / pointLightRadius[i] / 1.5f);

         //eyeballed to match fixed function appearance
         ratio = ratio * ratio * ratio * 0.4f; 

         pointLightTotal.x += ratio * lightDot * pointLightColorYCbCr.x / (baseLightColor.x + pointLightTotal.x + 1.5f);

         pointLightTotal.y += ratio * lightDot * pointLightColorYCbCr.y;
         pointLightTotal.z += ratio * lightDot * pointLightColorYCbCr.z;
      }
   }

   //pointLightTotal.x = clamp(pointLightTotal.x, 0.0f, 1.0f);
   pointLightTotal = colorYCbCrtoRGB(pointLightTotal);
   return pointLightTotal;
}

// Combine specular and direct lighting terms.
// note: if we make combinedColor "out" only, it throws a potentially uninitialized value warning, so we've made it inout
void applyLighting(inout vec4 combinedColor, vec4 albedo, float occlusionFactor)
{
   //large normal means glowing object
   if(worldNormal.x + worldNormal.y + worldNormal.z > 2.0f)
   {
      combinedColor = albedo;
      return;
   }  

   vec4 dirLightSpecular = occlusionFactor * specular * dirLightColor;
   dirLightSpecular *= 0.5f; //arbitrary adjustment
   vec4 dirLightDirect = ((NdotL * dirLightColor) * occlusionFactor) + (dirLightAmbient * occlusionFactor) + (dirShadowColor * (1.0f - occlusionFactor));

   if(NdotL <= 0.04f)
   {
      dirLightDirect = dirShadowColor;
      dirLightSpecular = vec4(0.0f, 0.0f, 0.0f, 0.0f);
   }
   else if(NdotL <= 0.1)
   {
      float val = (NdotL - 0.04f) / (0.1f - 0.04f);
      dirLightDirect = (dirLightDirect * val) + (dirShadowColor * (1.0f - val));
      dirLightSpecular = dirLightSpecular * val;
   }

   vec4 plt = accumulatePointLights(dirLightDirect); //point light total
   dirLightDirect += plt;

   dirLightSpecular.a = length(dirLightSpecular.rgb);
   dirLightDirect.a *= min(occlusionFactor + 0.75f, 1.0f);
   combinedColor.rgb = dirLightDirect.rgb * albedo.rgb;
   combinedColor.a = albedo.a;
   combinedColor += dirLightSpecular;
}

float shadowCoef();

vec4 fragmentMain()
{
   vec4 albedo = vec4(0.0f, 0.0f, 0.0f, 0.0f);
   calculateSurface(gl_Color, albedo);

   float occlusionFactor = 0.0f; 
   if(NdotL > -0.01f)
   {
      if(shadowSplitCount <= 0)
         occlusionFactor = 1.0f;
      else
         occlusionFactor = shadowCoef();
   }

   // Apply lighting and fog.
   vec4 fragColor = vec4(0.0f, 0.0f, 0.0f, 0.0f);
   if(isParticle == 1)
   {
      vec4 texAlbedo = texture2D(tex, texCoord);
      if(particleGlow == 1)
      {
         fragColor = texAlbedo * gl_Color;
         applyFog(fragColor);
         fragColor.a = texAlbedo.a * gl_Color.a;
      }
      else
      {
         vec4 dirLightDirect = (dirLightColor * occlusionFactor) + (dirLightAmbient * occlusionFactor) + (dirShadowColor * (1.0f - occlusionFactor));
         vec4 plt = accumulatePointLights(dirLightDirect);

         vec4 lightTotal = dirLightDirect + plt;
         lightTotal.x = clamp(lightTotal.x, 0.0f, 1.2f);
         lightTotal.y = clamp(lightTotal.y, 0.0f, 1.2f);
         lightTotal.z = clamp(lightTotal.z, 0.0f, 1.2f);
         
         fragColor = texAlbedo * gl_Color * lightTotal;

         applyFog(fragColor);
         fragColor.a = texAlbedo.a * gl_Color.a;
      }
   }
   else
   {
      applyLighting(fragColor, albedo, occlusionFactor);
      applyFog(fragColor);
   }
   
   // Uncomment to viz depth in B.
   //fragColor.z = vPos.y * 0.01f;

   return fragColor;
}